unit TBitClassUnit;
{$MODE DELPHI}
(**
     Code can be freely copied and used but not sold
     Original Author: minesadorada

   TBitClass is an object for manipulating binary arrays

   It has METHODS (private and public) and PROPERTIES (private and public)
   Note: PUBLIC METHODS USE NO PARAMETERS

   PRIVATE methods are used internally to this class and cannot be accessed
   via an object declaration

   All interaction with this code is via PUBLIC methods and properties.
   A pascal OOP class should be independent and self-contained.

   Any app should be able to put TBitClassUnit into it's interface section
   and interact via setting/getting properties and calling methods.

   It is created as an object, so needs a variable of type TBitClass to be declared
   and assigned/created.

   example:
   Var T20:TBitClass;

   Then used via properties and methods using the .(dot) notation
   example code:

   T20:=TBitClass.Create; //Create our TBitClass object and assign it to variable T20
   // Call InitOK method to initialise it
   If NOT T20.InitOK then
      begin
           raise ETBitClassException.Create(T20.ErrorString);
           T20.Free;
           Application.Terminate;
      end;
   // Set some properties
   T20.Number:=0;
   T20.Bit:=2; // Set the Bit property
   // Call a method that sets bit 2 of T20.Number
   // The method relies on the T20.Bit property having previously been set
   T20.SwitchBitOn;

   // Set the property to a different value
   T20.Bit:=1;
   b_truefalse=T20.IsBitOn; //Should return FALSE since only Bit 2 is on

   // Set the property to a different value
   T20.Bit:=2;
   b_truefalse=T20.IsBitOn; //Should return TRUE since Bit 2 is on

   T20.Free; //Always free an assigned object when done with it

   //Of course this class can do more than read and write bits!

   // Note: The Obj.ErrorString property can be read at any time
      for debugging purposes.  I find coding-in this mechanism
      very useful when developing a class.
**)
interface

Uses  Sysutils, Messages, Variants, Classes,Math,Dialogs;

Type
 ETBitClassException = class(Exception);
// To use from a form that does ObjTBitClass:=TBitClass.Create
// then set Errorstring and
// raise ETBitClassException.Create(ObjTBitClass.ErrorString);

CONST C_MAXBITS = 30; // Used as error test in loads of routines
   // This class can handle numbers 31 bits wide!

TYPE TArrayofCardinal = array of Cardinal; // Used for fcUnSureNumbersArray

TYPE
 TBitClass = class(TObject)
 private
  // PRIVATE VARS - Unavailable outside this class
  fErrorString:String; // Property 'ErrorString'
  fiBit:Integer; // Property 'Bit'
  fcNumber:Cardinal; // Property 'Number'
  fbBitArray:array[0..C_MAXBITS] of Boolean;
 // Unsure vars
  fbUnsureBitArray:array[0..C_MAXBITS] of Boolean;
  fiUnsureBit:Integer;
  fcUnSureNumbersArray:TArrayofCardinal;
  fMaxBits:Integer;

  // PRIVATE METHODS - Unavailable outside this class
  Function Priv_IsBitOn(cNumber:Cardinal;iBit:Integer):Boolean;
  Function Priv_IsBitOff(cNumber:Cardinal;iBit:Integer):Boolean;
  Function Priv_BitOn(cNumber:Cardinal;iBit:Integer):Cardinal;
  Function Priv_BitOff(cNumber:Cardinal;iBit:Integer):Cardinal;

  // PRIVATE PROPERTY GET/SET ROUTINES
  Procedure SetErrorString(aString:String); // Sets 'Errorstring' property
  Procedure SetBit(iBit:Integer); // Sets 'Bit" property
  Procedure SetNumber(cNum:Cardinal); // Sets 'Number' property
  Function GetNumberBitString:String; // Sets 'NumberBitString' property
  Function GetBitBitString:String; // Sets 'BitBitString' property
  // Unsure Methods
  Procedure SetUnsureBit(iUnsureBit:Integer);

 protected
  // This can be overridden in a descendant
   Function InitObj:boolean; virtual;

 public
 // PUBLIC METHODS
   // NOTE NO PARAMETERS - CLASSES USE PUBLIC PROPERTIES INSTEAD
   constructor Create;
   destructor Destroy; override;
   // Call as MyInstance:=TBitClass.Create;
   Procedure ShowAbout; virtual; // Lists Methods and Properties
   Function InitOK:Boolean; // If InitOK then proceed...
   Procedure ResetAll; // Use to reset all numbers, arrays and bits
   Function Power2(exp:Integer):Cardinal;
   Function BitString(cNum:Cardinal):String; // For display
   Function IsBitOn:Boolean; // Uses preset Number and Bit property values
   Function IsBitOff:Boolean; // Uses preset Number and Bit property values
   Procedure SwitchBitOn; // Uses preset Number and Bit property values
   Procedure SwitchBitOff; // Uses preset Number and Bit property values
   // Used for testing
   Function GetNumberFromBitArray:Cardinal;
   Function GetNumberFromUnsureBitArray:Cardinal;
// Unsure Methods
   Procedure SwitchUnsureBitOn; // Affects an internal boolean array
   Procedure SwitchUnsureBitOff; // Affects an internal boolean array
   Procedure GetUnsureNumberArray; // Sets the property UnsureNumberArray:TArrayOfCardinal
 Published
 // PUBLIC PROPERTIES  - available via Obj.Property:=SomeVar
 //                             or SomeVar:=Obj.Property
   // Some read-only, some read + write
   Property ErrorString:String read fErrorString write SetErrorString;
   Property Bit:Integer read fiBit write SetBit;
   Property Number:Cardinal read fcNumber write SetNumber;
   Property NumberBitString:String read GetNumberBitString;
   Property BitBitString:String read GetBitBitString;
   Property UnsureBit:Integer read fiUnsureBit Write SetUnsureBit;
   Property MaxBits:Integer read fMaxBits;
   // NOTE THE RETURN TYPE OF THIS PROPERTY
   // Declare a Var of type TArrayCardinal before reading this property
   Property UnsureNumberArray:TArrayOfCardinal read fcUnSureNumbersArray;
end;



implementation
// *********************************************************************
constructor TBitClass.Create;
begin
 inherited Create;
end;
destructor TBitClass.Destroy;
begin
 inherited Destroy;
 fcUnSureNumbersArray:=Nil;
end;
// *********************************************************************
Procedure TBitClass.ShowAbout;
CONST CR = #10+#13;
Var s:String;
begin
 s:='==== TBitClass Object ====' + CR;
 s:=s + 'Properties:' + CR;
 s:=s + '1) Bit:Integer range 0..' + IntToStr(C_MAXBITS) + CR;
 s:=s + '2) Number:Cardinal range 0..' + IntToStr(C_MAXBITS) + '^2' + CR;
 s:=s + '3) NumberBitString and BitBitString' + CR;
 s:=s + '4) ErrorString:String' + CR;
 s:=s + 'Methods:' + CR;
 s:=s + '1) Power2(exp:Integer):Cardinal' + CR;
 s:=s + 'or Power2(exp:extended):Extended' + CR;
 s:=s + '2) SwitchBitOn and SwitchBitOff' + CR;
 s:=s + '3) IsBitOn and IsBitOff' + CR;
 s:=s + '4) BitString:String' + CR;
 MessageDlg(s,mtInformation,[MBOK],0);
end;

// 2 Versions of Power2 Method
// Call with Extended or Integer parameter
// Extended version commented out as unnecessary
(**
Function TBitClass.Power2(exp:Extended):Extended;
 begin
   Result:=Power(2,Exp);
end;
**)
Function TBitClass.Power2(exp:Integer):Cardinal;
// TBitClass.Power2(bits)
// TBitClass.Power2(8)=256
Const Base = 2;
Var f:Integer;
begin
// DEBUGGING
// ShowMessage(ErrorString + ':' + IntToStr(Exp));
ErrorString:=Format('TBitClass.Power2 Exp=%d',[exp]);
Result:=1;
If (Exp > C_MAXBITS) or (Exp < 0) then
 begin
   raise ETBitClassException.CreateFmt
   (ErrorString + ': Exponent not within the valid range of 0..%d',[C_MAXBITS]);
   Result:=0;
 end;
// Do the calculation
For f:=0 to exp-1 do
 Result:=Result * base;
end;
// *********************************************************************
Function TBitClass.Priv_IsBitOn(cNumber:Cardinal;iBit:Integer):Boolean;
Var cBit:Cardinal;
Begin
  ErrorString:=Format('TBitClass.Priv_IsBitOn Number=%d, iBit=%d',[cNumber,iBit]);
  if (iBit < 0) OR (iBit > C_MAXBITS) then ETBitClassException.CreateFmt(
    ErrorString + ' Bit "%d" is not within the valid range of 0..%d',[iBit,C_MAXBITS]);
  cBit:=Power2(iBit);
  If (cNumber AND cBit) = cBit then Result:=True
  else Result:=False;
End;
// *********************************************************************
Function TBitClass.Priv_IsBitOff(cNumber:Cardinal;iBit:Integer):Boolean;
Var cBit:Cardinal;
Begin
  ErrorString:=Format('TBitClass.Priv_IsBitOff Number=%d, iBit=%d',[cNumber,iBit]);
  if (iBit < 0) OR (iBit > C_MAXBITS) then ETBitClassException.CreateFmt(
    ErrorString + ' Bit "%d" is not within the valid range of 0..%d',[iBit,C_MAXBITS]);
  cBit:=Power2(iBit); // Fetch Binary
  If (cNumber AND cBit) = 0 then Result:=True
  else Result:=False;
End;
// *********************************************************************
Function TBitClass.IsBitOn:Boolean;
// PUBLIC METHOD
Begin
   ErrorString:='TBitClass.IsBitOn';
   Result:=Priv_IsBitOn(fcNumber,fiBit);
End;
// *********************************************************************
Function TBitClass.IsBitOff:Boolean;
// PUBLIC METHOD
Begin
   Result:=Priv_IsBitOff(fcNumber,fiBit);
End;
// *********************************************************************
Function TBitClass.Priv_BitOn(cNumber:Cardinal;iBit:Integer):Cardinal;
Var cBit:Cardinal;
// Private Method
begin
 ErrorString:=Format('TBitClass.BitOn Number=%d, iBit=%d',[cNumber,iBit]);
 Result:=0;
 if (iBit < 0) OR (iBit > C_MAXBITS) then ETBitClassException.CreateFmt(
  ErrorString + ' Bit "%d" is not within the valid range of 0..%d',[iBit,C_MAXBITS]);
 cBit:=Power2(iBit); // Fetch Binary
 Result:=cNumber OR cBit;
end;
// *********************************************************************
Function TBitClass.Priv_BitOff(cNumber:Cardinal;iBit:Integer):Cardinal;
Var cBit:Cardinal;
// Private Method
begin
 ErrorString:=Format('TBitClass.BitOff Number=%d, iBit=%d',[cNumber,iBit]);
 Result:=0;
 if (iBit < 0) OR (iBit > C_MAXBITS) then ETBitClassException.CreateFmt(
  ErrorString + ' Bit "%d" is not within the valid range of 0..%d',[iBit,C_MAXBITS]);
 cBit:=Power2(iBit); // Fetch Binary
 cBit := NOT cBit; // Negate bitfield (now all 'I's except iBit)
 Result:=cNumber AND cBit; // Switch off
end;
// *********************************************************************
Procedure TBitClass.SwitchBitOn;
// PUBLIC METHOD
begin
 ErrorString:='TBitClass.SwichBitOn';
 fcNumber:=Priv_BitOn(fcNumber,fiBit);
 fbBitArray[fiBit]:=TRUE;
end;
// *********************************************************************
Procedure TBitClass.SwitchBitOff;
// PUBLIC METHOD
begin
 ErrorString:='TBitClass.SwichBitOff';
 fcNumber:=Priv_BitOff(fcNumber,fiBit);
 fbBitArray[fiBit]:=FALSE;
end;
// *********************************************************************
Function TBitClass.BitString(cNum:Cardinal):String;
// Converts a number (< 31 bits) into
// A string of form 00110011 etc
// and trims off leading zeros
// Returns '0' if iNum is Zero
Var
 f:Integer;
 s:String;
begin
ErrorString:=Format('TBitClass.BitString Number=%d',[cNum]);
Result:='0';
//CrapChecks
if (cNum > Power2(C_MAXBITS)) then
  begin
    ETBitClassException.CreateFmt(
     ErrorString + ' Bit "%d" is not within the valid range of 0..%d',[cNum,C_MAXBITS]);
    exit;
  end;

// 1) Convert into a complete string
For f:=0 to C_MAXBITS-1 do
  begin
   fiBit:=f;
   // Use private method
   If Priv_IsBitOn(cNum,f) then Result:='1' + Result
   else Result:='0' + Result;
  end;
  s:=Result;
// DEBUG
//  ShowMessage('BitString' + ' ' + Result + ' ' + IntToStr(fcNumber));

// 2) Trim off leading zeros
For f:=1 to Length(Result) do
   If Result[f] ='1' then
    begin
     	Result:=Copy(Result,f,Length(result) - f);
     	Break; // Return trimmed result
    end;
// If no '1's are found - just return default '0'
end;
// *********************************************************************
Function TBitClass.InitObj:boolean;
 begin
   ErrorString:='TBitClass.InitObj';
   ResetAll; // Set all bits and arrays etc to 0
   Result:=True;
 end;
// *********************************************************************
Function TBitClass.InitOK:Boolean;
 begin
   ErrorString:='TBitClass.InitOK';
 // Call after Create.
 // Any problems, set ErrorString and call the ETBitClassException
   Result:=InitObj;
 end;

// *********************************************************************
// GET/SETS
Procedure TBitClass.SetErrorString(aString:String);
begin fErrorString:=aString; end;
// *********************************************************************
Procedure TBitClass.SetBit(iBit:Integer);
begin
  ErrorString:=Format('TBitClass.SetBit Bit=%d',[iBit]);
  if fiBit <> iBit then
    begin
  	if (iBit < 0) OR (iBit > C_MAXBITS) then ETBitClassException.CreateFmt(
  	ErrorString + ' Bit "%d" is not within the valid range of 0..%d',[iBit,C_MAXBITS])
  	else fiBit:=iBit;
    end;
end;
// *********************************************************************
Procedure TBitClass.SetNumber(cNum:Cardinal);
Var cTooBig:Cardinal;
begin
  ErrorString:=Format('TBitClass.SetNumber Number=',[cNum]);
  If cNum <> fcNumber then
    begin
      fcNumber:=0; // default
      cToobig:=Power2(C_MAXBITS);
      If (cNum > cTooBig) then
      ETBitClassException.CreateFmt(
 	ErrorString + ' Number "%d" is not within the valid range of 0..%d',[cNum,cTooBig])
      else fcNumber:=cNum;
    end;
end;
// *********************************************************************
Function TBitClass.GetNumberBitString:String;
Begin
  ErrorString:='TBitClass.GetNumberBitString';
  Result:=BitString(fcNumber);
End;
// *********************************************************************
Function TBitClass.GetBitBitString:String;
// Fetches binary string of fiBit
Var cBit:Cardinal;
Begin
  ErrorString:='TBitClass.GetBitBitString';
  cBit:=Power2(fiBit);
  Result:=BitString(cBit);
End;
// *********************************************************************
Procedure TBitClass.SetUnsureBit(iUnsureBit:Integer);
begin
  ErrorString:='TBitClass.SetUnsureBit';
  If iUnsureBit <> fiUnsureBit then
    If (iUnsureBit < 0) OR (iUnsureBit > C_MAXBITS) then
   	ETBitClassException.CreateFmt(
   	ErrorString + ' Unsure bit "%d" is not within the valid range of 0..%d',
   	[iUnsureBit,C_MAXBITS])
    else
   	fiUnsureBit:=iUnsureBit;

end;
// *********************************************************************
Procedure TBitClass.SwitchUnsureBitOn;
// PUBLIC METHOD
// USES ARRAY
begin
  ErrorString:=Format('TBitClass.SwitchUnsureBitOn Bit=%d',[fiUnsureBit]);
  fbUnsureBitArray[fiUnsureBit]:=TRUE;
end;
// *********************************************************************
Procedure TBitClass.SwitchUnsureBitOff;
// PUBLIC METHOD
// USES ARRAY
begin
  ErrorString:=Format('TBitClass.SwitchUnsureBitOff Bit=%d',[fiUnsureBit]);
  fbUnsureBitArray[fiUnsureBit]:=FALSE;
end;
// *********************************************************************
Procedure TBitClass.ResetAll;
// PUBLIC METHOD
Var f:Integer;
begin
  ErrorString:='TBitClass.ResetAll';
  For f:=0 to C_MAXBITS do fbUnsureBitArray[f]:=FALSE;
  For f:=0 to C_MAXBITS do fbBitArray[f]:=FALSE;
  fiBit:=0;
  fcNumber:=0;
  fcUnSureNumbersArray:=Nil;
  fMaxBits:=C_MAXBITS;
end;
// *********************************************************************
Procedure TBitClass.GetUnsureNumberArray;
// PUBLIC METHOD
// Sets the property UnsureNumberArray:TArrayOfCardinal
// This is a zero-base open array
// The highest index is High(UnsureNumberArray)


// Using a combination of Bits and UnsureBits arrays of bool.
// Fetch all the combinations of numbers and put them into an array of cardinals
// fcNumber has already been set.
Var fCount,gCount:Integer;
    cCombinations:Cardinal; // Power2 of the number of ON bits in the Unsure Bitarray
    iCombinationsArray:array of Integer; // Bit positions of TRUE's in fbUnsureBitArray
    iCombinationsArrayCount:Integer;
    iUnsureNumberArrayCount:Integer;
begin
  ErrorString:='TBitClass.GetUnsureNumberArray';
  iUnsureNumberArrayCount:=0;
  iCombinationsArrayCount:=0;
  cCombinations:=0;
// Find the total combinations (Number of ON bits)^2
// The sequence 0 - this number gives ALL combinations of bits
// i.e. 8: 000=0 001=1 010=2 011=3 100=4 101=5 110=6 111=7
  For fCount:=0 to High(fbUnsureBitArray) do
    If fbUnsureBitArray[fCount]=TRUE then
     begin
  	// ShowMessage(format('Bit %d is ON',[fCount]));
   	Inc(cCombinations);
        Inc(iCombinationsArrayCount);
        SetLength(iCombinationsArray,iCombinationsArrayCount);
        // These are the BIT POSITIONS in an array (ZERO BASED)
        iCombinationsArray[iCombinationsArrayCount-1]:=fCount;
     end;

//   ShowMessage(format('No. of members in iCombinationsArray = %d',[iCombinationsArrayCount]));


  cCombinations:=Power2(cCombinations);
  // This gives the total number
  // ShowMessage(format('Power 2 Combinations = %d',[cCombinations]));

  // Make it zero-base (0..Total-1)
  Dec(cCombinations);

// The binary sequence of 0..cCombinations will give
// us all possible combinations of the ON/OFF
// sequence of bits
// If cCombinations=1 (01) then 0=0,1=1
// If cCombinations=2 (10) then 00=0,01=1,10=2
// If cCombinations=7 (111) then 000=0 001=1 010=2 011=3 100=4 101=5 110=6 111=7

// We now need to map these bits to the BitArray and extract the numbers
// The mapping will go
// Bit 0 to the lowest TRUE position
// Bit 1 to the second lowest TRUE position
// Bit 3 to the third lowest TRUE position
// ..and so on up to the number of bits in cCombinations

// 2) Make up the numbers array to AND with fcNumber
// iCombinationsArray[Low..High] contains the bit position
//  numbers in fbUnsureBitArray


  For fCount:=0 to cCombinations do
   begin
     for gCount:=0 to High(iCombinationsArray) do
      begin
       If Priv_IsBitOn(fCount,gCount) then
          fbBitArray[iCombinationsArray[gCount]]:=TRUE
        else
          fbBitArray[iCombinationsArray[gCount]]:=FALSE;
      end; // of gCount..
          Inc(iUnsureNumberArrayCount);
          SetLength(fcUnSureNumbersArray,iUnsureNumberArrayCount);
          fcUnSureNumbersArray[iUnsureNumberArrayCount-1]:=GetNumberFromBitArray;
   end; // of fCount

end;
// *********************************************************************
Function TBitClass.GetNumberFromBitArray:Cardinal;
// PUBLIC METHOD
// fbBitArray:array[0..C_MAXBITS] of Boolean;
// 0 = LSB
Var fCount:Integer;
    cBit:Cardinal;
begin
  ErrorString:='TBitClass.GetNumberFromBitArray';
  Result:=0;
  For fCount:=0 to High(fbBitArray) do
   begin
    If fbBitArray[fCount]=TRUE then
     Begin
    	  cBit:=Power2(fCount); // Fetch Binary
    	  Result:=Result OR cBit;
     end; // of if..
   end;// of for..
end;
// *********************************************************************
Function TBitClass.GetNumberFromUnsureBitArray:Cardinal;
// PUBLIC METHOD
// fbUnsureBitArray:array[0..C_MAXBITS] of Boolean;
// 0=LSB
Var fCount:Integer;
    cBit:Cardinal;
begin
  ErrorString:='TBitClass.GetNumberFromUnsureBitArray';
  Result:=0;
  For fCount:=0 to High(fbUnsureBitArray) do
   begin
    If fbUnsureBitArray[fCount]=TRUE then
     Begin
    	cBit:=Power2(fCount); // Fetch Binary
    	Result:=Result OR cBit;
     end; // of if..
   end;// of for..
end;
// *********************************************************************
// *********************************************************************
// *********************************************************************
end.


