unit TFileArrayUnit;
{$MODE DELPHI}
(*
     Code can be freely copied and used but not sold
     Original Author: minesadorada

  This is an object that handles a string array of arrays
  Each array element (row) has an associated array (cols)

  Each row can have a variable number of columns, so that
  the array is quite efficient in memory and disk space

  It can save and load the array to disk

  It was designed to work with the TBitClass object in
  a 'twenty questions' application.  The array stores
  all the data in a logically searchable form.

  Default maximum operation is 2^19 rows with each row having 255 columns
  In the app, each row represents a primary 'category' with up to 254 alternatives
  under the same category definition.

  Cardinal numbers are used to keep platform independence.

*)
interface
Uses  Sysutils, Variants, Classes,Dialogs,Controls,FileUtil;
CONST C_MAXARRAY = 31; // = MAX bits
CONST C_FILENAME='filearray.dat'; // Default Data file name
CONST C_MAXELEMENTLENGTH = 255;
CONST C_SUREMARKER = '*';

Type
 ETFileArrayException = class(Exception);
Type TGuess = array of array of string;
// To allocate:
//1 ) SetLength(Guess,rows,columns)
//2 ) SetLength(Guess,rows)
//3 ) SetLength(Guess[0],3) then SetLength(Guess[1],4) two unequal rows
// To iterate: for f:=Low(Guess) to High(guess) do
// 		  for g:=Low(Guess[f]) to High(Guess[f])
//			Guess[f,g] = avalue

Type TRecGuess = Record
 Row:Cardinal; // Number
 Col:Cardinal; // Entry in Number
 Element:ShortString; // Fixed length because of record format
end;
// This record type is used to serialise the data
// It is in a simpler format to the guess array, so has to
// hold the postion of the element in the array as seperate
// Row and Column fields.  In that way, it can be read from
// disk and each Element allocated correctly to the array.
// To save the array directly to disk would be very wasteful of space.

// A File of records is easy to handle.
TYPE tFileRecGuess = File of TRecGuess; // File storage format

TYPE
 TFileArray = class(TObject)
 private
  // Private vars are used to read and write public properties.
 fRow:Integer; // Guess[fCol]
 fCol:Integer; // Guess[fRow,fCol]
 fMaxArray:Integer; // Maximum Number allowed
 fDatFilePath:String; // Always current directory
 fRecs:Integer; // Total No. of entries
 Guess:TGuess; // Dynamic array holds all data
 fbUseUnsure:Boolean; // UseUnsure property
 fCharSureMarker:Char; // SureMarker property
 fbSure:Boolean; // Sure property
 fbDataSaved:Boolean; // Proprty: Updated internally
 fErrorString:String;
 fcFoundRow:Cardinal; // Used as result for SearchForElement
 fcFoundCol:Cardinal; // Used as result for SearchForElement
 fUniqueElementList:TStringList; // myList.Assign(TFileArray.GetUniqueElementList);
 // or combocontrol.Items:=TFileArray.GetUniqueElementList
 fFindElementList:TStringList; // myList.Assign(TFileArray.GetElementListFromNumber)
 // or combocontrol.Items:=TFileArray.GetElementListFromNumber

  Function Power2(exp:Integer):Cardinal;
  Procedure SetRow(iValue:Integer);
  Procedure SetCol(iValue:Integer);
  Procedure SetUseUnsure(bValue:Boolean);
  Procedure SetSureMarker(charValue:Char);
  Procedure SetSure(bValue:Boolean);
  Function GetElementMaxCol:Integer;
  Procedure SetErrorString(aString:String);
  Procedure SetDatFilePath(aString:String);
  Function Priv_DeleteSubElement(aRow:Cardinal;aCol:Integer):boolean;//  Function GetMaxArray:Integer;
  Function Priv_DeleteElement(aRow:Cardinal):boolean;
 protected
   Function InitObj:boolean;
 public
  // PUBLIC METHODS
  constructor Create(iMaxBits:Integer);
  destructor Destroy;override;
  Function InitOK:boolean;
  Procedure SetElement(sValue:String);
  Function GetElement:String;
  Function SaveToFile:Boolean;
  Function LoadFromFile:Boolean;
  Procedure DeleteDataFile;
  Function SearchForElement(sValue:String):Boolean;
  Function DeleteAllInstancesOf(sValue:String;Delimiter:String):Integer;
  Procedure DeleteSubElement; // Deletes an entry in a row
  Procedure DeleteElement; // Deletes a whole row
  Function DeleteData:Boolean;
  Function StripSureMarker(sValue:String):String;

  Function GetUniqueElementList:TStringList;
  // myList.Assign(TFileArray.GetUniqueElementList)
  // or myCombobox.Items:=TFileArray.GetUniqueElementList
  Function GetElementListFromNumber:TStringList;
  // myList.Assign(TFileArray.GetElementListFromNumber)
  // or myCombobox.Items:=TFileArray.GetElementListFromNumber

 published
  // Read + Write properties
  Property Row:Integer read fRow write SetRow;
  Property Col:Integer read fCol write SetCol;
  Property DataFilePath:String read fDatFilePath write SetDatFilePath;
  Property UseUnsure:Boolean read fbUseUnsure write SetUseUnsure;
  Property SureMarker:Char read fCharSureMarker write SetSureMarker;
  Property Sure:Boolean read fbSure write SetSure;
  Property ErrorString:String read fErrorString write SetErrorString;
  // Read-only Properties
  Property Recs:Integer read fRecs;
  Property MaxCol:Integer read GetElementMaxCol;
  Property MaxBits:Integer read fMaxArray;
  Property DataSaved:Boolean read fbDataSaved;
  Property FoundRow:Cardinal read fcFoundRow;
  Property FoundCol:Cardinal read fcFoundCol;
end; // of Type

implementation
Constructor TFileArray.Create(iMaxBits:Integer);
begin
//  Inherited Create;

// Initialise default values and create sub-objects here
// Note that the parameter iMaxBits sets the size of the array.
// When working with TBitClassUnit, the BitSize must agree!

  fMaxArray:=iMaxBits+1; // Convert from zero-based to 1-based.
// Deal with unrealistic parameter if present
  if fMaxArray < 0 then fMaxArray:=1;
  If fMaxArray > C_MAXARRAY then fMaxArray:=C_MAXARRAY;

// Set the array size.  If bitsize=8, then Rows=256 etc.
  SetLength(Guess,Power2(fMaxArray));
  fbDataSaved:=False; // Set dirty flag
  // Now set up the temporary string lists
  fUniqueElementList:=TStringList.Create;
  fUniqueElementList.Sorted:=True;
  fUniqueElementList.Duplicates:=dupIgnore;


  fFindElementList:=TStringList.Create;
  fFindElementList.Sorted:=True;
  fFindElementList.Duplicates:=dupIgnore;


end;
// ***************************************************************************//
Destructor TFileArray.Destroy;
// PUBLIC METHOD
begin
  Guess:=Nil;// Very important to free the massive array from memory!
  // fUniqueElementList.Free;
  // fFindElementList.Free;
end;
// ***************************************************************************//
Function TFileArray.DeleteData:Boolean;
begin
 // Re-initialise everything and delete all data
 SetLength(Guess,0);
 fRecs:=0;
 fbDataSaved:=True; // Don't resave
 DeleteDataFile;
 Result:=True;
end;
// ***************************************************************************//
Function TFileArray.StripSureMarker(sValue:String):String;
// PUBLIC METHOD

// The 'sure marker' denotes a subelement that has preference
begin
  ErrorString:=Format('TFileArray.StripSureMarker String=%s',[sValue]);
  If (Length(sValue) > 1) then
    If sValue[1] = fCharSureMarker then
      sValue:=Copy(sValue,2,Length(sValue)-1);
  Result:=sValue
end;
// ***************************************************************************//
Function TFileArray.GetUniqueElementList:TStringList;
// PUBLIC METHOD
// Returns Sorted stringlist with no duplicates
Var
    BigLoop,SmallLoop:Cardinal;

begin
// if fUniqueElementList.Count > 0 then fUniqueElementList.Clear;
 Result:=fUniqueElementList; // Assign to empty pre-initialised list
 ErrorString:='TFileArray.GetUniqueElementList';
 If High(Guess)=-1 then Exit;
  For BigLoop:=Low(Guess) to High(Guess) do
     For SmallLoop:=Low(Guess[BigLoop]) to High(Guess[BigLoop]) do
       BEGIN // Iterate through all Guess entries
        If High(Guess[BigLoop])=-1 then Break;
        Result.Add(StripSureMarker(Guess[BigLoop,SmallLoop]));
       END;
end;
// ***************************************************************************//
Function TFileArray.Priv_DeleteElement(aRow:Cardinal):boolean;
Var i:Integer;
begin
  ErrorString:=Format('TFileArray.Priv_DeleteElement Number=%d',[aRow]);
  i:=High(Guess[aRow])+1; // Convert zero-base to 1-base
  SetLength(Guess[aRow],0);
  fbDataSaved:=False;
  Dec(fRecs,i);
  Result:=True;
end;
// ***************************************************************************//
Procedure TFileArray.DeleteElement;
// PUBLIC METHOD
// Calls private method to do the deed.
begin
 ErrorString:=Format('TFileArray.DeleteElement Number=%d',[fRow]);
 Priv_DeleteElement(fRow);
end;
// ***************************************************************************//
Function TFileArray.Priv_DeleteSubElement(aRow:Cardinal;aCol:Integer):boolean;
Var fCount:Cardinal;
begin
 ErrorString:=Format('TFileArray.Priv_DeleteSubElement Number=%d, Element=%d',[aRow,aCol]);
 Result:=False;
 If High(Guess[fRow])=-1 then Exit; // No entry at fCol,fRow

 // Now the deletion..
 IF aCol=0 then // Only one entry for this BigLoop
 BEgin
    Result:=Priv_DeleteElement(aRow);
    Exit;
   // SetLength(Guess[aRow,aCol],0)
 ENd
 ELse
   BEgin
    // aCol > 0
    If aCol=High(Guess[aRow]) then
    // Entry is last one in list
      begin
        // Set New Length
        SetLength(Guess[aRow,aCol],High(Guess[aRow])-1);
      end
      Else
      // aCol > 0 AND aCol < High(Guess[aRow)
      // Entry is in middle of list
      begin
        For fCount:=aCol to High(Guess[aRow])-1 do
          begin
          // Shuffle entries down
           Guess[aRow,fCount]:=Guess[aRow,fCount+1];
          end;
        SetLength(Guess[aRow,aCol],High(Guess[aRow])-1);
      end; //.. of aCol=High(Guess..
    ENd; // //of aCol=0

 fbDataSaved:=False;
 Dec(fRecs);
 Result:=True;
end;
// ***************************************************************************//
Procedure TFileArray.DeleteSubElement;
// PUBLIC METHOD
//Uses fRow,fCol
begin
 ErrorString:=Format('TFileArray.DeleteSubElement Number=%d, Element=%d',[fRow,fCol]);
  // Call private function
  Priv_DeleteSubElement(fRow,fCol);
end;
// ***************************************************************************//
Function TFileArray.DeleteAllInstancesOf(sValue:String;Delimiter:String):Integer;
// PUBLIC METHOD
// Loop through the whole Guess array
// Deleting all matches for sValue as they are found
Var
  BigLoop,SmallLoop:Cardinal;
  s:String;
  i:Integer;
begin
Result:=0;
 ErrorString:=Format('TFileArray.DeleteAllInstancesOf Value=%s',[sValue]);
  For BigLoop:=Low(Guess) to High(Guess) do
     For SmallLoop:=Low(Guess[BigLoop]) to High(Guess[BigLoop]) do
       BEGIN // Iterate through all Guess entries
       If High(Guess[BigLoop])=-1 then Break;
       s:=Guess[BigLoop,SmallLoop];
       if Delimiter <> '' then
         begin
          i:=Pos(Delimiter,s);
          if i > 0 then s:=LeftStr(s,i-1);
         end;

         If (CompareText(s,sValue)=0) or (CompareText(s,fCharSureMarker + sValue)=0) then
           begin // We have a match!
           // ShowMessage(Format('Match "%s" at Row %d, Column %d',[s,BigLoop,SmallLoop]));
             Inc(Result);
             fbDataSaved:=False;
	     //Delete it!
             // [BigLoop,SmallLoop] gives index
             Priv_DeleteSubElement(bigLoop,SmallLoop);
           end; // ..of CompareText=0
       END; // ..of For SmallLoop
Dec(fRecs,Result);
end;
// ***************************************************************************//
Function TFileArray.GetElementListFromNumber:TStringList;
// PUBLIC METHOD

Var ColumnCount:Integer;
Begin
ErrorString:='TFileArray.GetElementListFromNumber';
fFindElementList.Clear;
Result:=fFindElementList; // Assign empty list to return value
  For ColumnCount:=Low(Guess[fRow]) to High(Guess[fRow]) do
    begin
//     ShowMessage(Format('Looking at Row %d, Column %d',[fRow,ColumnCount]));
     If Length(Guess[fRow,ColumnCount]) > 0 then
        Result.Add(Guess[fRow,ColumnCount]);
    end;
end;
// ***************************************************************************//
Function TFileArray.SearchForElement(sValue:String):Boolean;
// PUBLIC METHOD
// Sets readonly properties FoundRow, FoundCol and Sure
// Finds only first instance
Var RowCount,ColumnCount:Integer;
s:String;
begin
 ErrorString:=Format('TFileArray.SearchForElement Value=%s',[sValue]);
 Result:=False;
 fcFoundRow:=0;
 fcFoundCol:=0;
 for RowCount:=Low(Guess) to High(guess) do
  Begin
    For ColumnCount:=Low(Guess[RowCount]) to High(Guess[RowCount]) do
      begin
        s:=Guess[RowCount,ColumnCount];
        // Set SURE Property and strip out marker
        If Length(s) > 1 then
           If s[1]=fCharSureMarker then
             fbSure:=TRUE
           else
            fbSure:=FALSE;
        s:=Copy(s,2,Length(s)-1);
        If LowerCase(s) = LowerCase(sValue) then
          begin
            Result:=True;
            fcFoundRow:=RowCount;
            fcFoundCol:=ColumnCount;
            Break;
          end;
      end;
  End;

end;
// ***************************************************************************//
Procedure TFileArray.DeleteDataFile;
// PUBLIC METHOD
begin
  ErrorString:='TFileArray.DeleteDataFile';
  If FileExists(fDatFilePath) then DeleteFileUTF8(PChar(fDatFilePath));
  fbDataSaved:=False;
end;

// ***************************************************************************//
Function TFileArray.SaveToFile:Boolean;
// PUBLIC METHOD
Var fc,gc:Integer;
    rRecGuess:TRecGuess;
    F:tFileRecGuess;
    s:String;
    ss:ShortString;
begin
ErrorString:='TFileArray.SaveToFile';
Result:=False;
AssignFile(F,fDatFilePath);
TRY
// Make backup first!
If FileExistsUTF8(fDatFilePath) then
CopyFile(PChar(fDatFilePath),PChar(ChangeFileExt(fDatFilePath,'.bak')),FALSE);
Rewrite(F); // Always erase existing file
FileMode := 1;  {Set file access to write only }
fRecs:=0;
   for fc:=Low(Guess) to High(guess) do
    begin
     	for gc:=Low(Guess[fc]) to High(Guess[fc]) do
         begin
           If Length(Guess[fc,gc]) > 0 then
             begin
           	rRecGuess.Row:=fc;
   	   	    rRecGuess.Col:=gc;
	        s:=Guess[fc,gc];
           	ss:=Copy(s,1,Length(s));
   	   	    rRecGuess.Element:=ss;
           	Write(F,rRecGuess);// Write to file
                Inc(fRecs);
             end;
         end;
    end;
FINALLY
CloseFile(F);
END;
if fRecs > 0 then
 begin
  Result:=True;
  fbDataSaved:=True;
 end;
//ShowMessage(Format('%d records written',[index]));
end;

// ***************************************************************************//
Function TFileArray.LoadFromFile:Boolean;
// PUBLIC METHOD
Var iRow,iCol:Integer;
    rRecGuess:TRecGuess;
    F:tFileRecGuess;
    sElement:String;
begin
ErrorString:='TFileArray.LoadFromFile';
Result:=False;
If NOT FileExists(fDatFilePath) then Exit;
AssignFile(F,fDatFilePath);
fRecs:=0;
FileMode := 0;  {Set file access to read only }
Reset(F);
TRY
  WHILE NOT EOF(f) do
    begin
    // Read into temp record
     Read(F,rRecGuess);
     iRow:=rRecGuess.Row;
     iCol:=rRecGuess.Col;
     sElement:=rRecGuess.Element;

     // Push contents of temp record into Guess array
      SetLength(Guess[iRow],iCol + 1);
      Guess[iRow,iCol]:=sElement;
     // Update Recs property
      Inc(fRecs);
    end;
    if fRecs > 0 then
     begin
      Result:=True;
      fbDataSaved:=True;
     end;
//    ShowMessage(Format('%d records read',[fRecs]));
FINALLY
CloseFile(F);
END;
end;
// ***************************************************************************//
Function TFileArray.Power2(exp:Integer):Cardinal;
Const Base = 2;
Var f:Integer;
begin
ErrorString:=Format('TFileArray.Power2 Exp=%d',[exp]);
Result:=1;
If (Exp > C_MAXARRAY) or (Exp < 0) then
 begin
   raise ETFilearrayException.CreateFmt
   ('(Cardinal) function called with Exp of %d',[Exp]);
   Result:=0;
 end;
// Do the calculation
For f:=0 to exp-1 do
 Result:=Result * base;
end;
// ***************************************************************************//
Function TFileArray.InitObj:boolean;
begin
  ErrorString:='TFileArray.InitObj';
  fDatFilePath:=ExtractFilePath(ParamStr(0)) + C_FILENAME;
  fRecs:=0;
  fbUseUnsure:=TRUE;
  fCharSureMarker:=C_SUREMARKER;
  fbSure:=FALSE;
  Result:=True;
  fbDataSaved:=False;
  fcFoundRow:=0;
  fcFoundCol:=0;
end;
// ***************************************************************************//
Function TFileArray.InitOK:boolean;
// PUBLIC METHOD
begin
  ErrorString:='TFileArray.InitOK';
  Result:=InitObj; // Calls private method
end;
// ***************************************************************************//
Procedure TFileArray.SetElement(sValue:String);
// PUBLIC METHOD
// Note - no parameters
Begin
 ErrorString:=Format('TFileArray.SetElement Value=%s',[sValue]);
 If (fRow > High(Guess)) then SetLength(Guess,fRow+1);
 If (fCol > High(Guess[fRow])) then SetLength(Guess[fRow],fCol+1);
// Debug
// ShowMessage(Format('Row = %d, Column = %d, High(Guess) = %d, High(Guess[row] = %d',
// [row,Column,High(Guess),High(Guess[row])]));

 // Mark SURE entries
 If (fbUseUnsure=TRUE) AND (fbSure=TRUE) then sValue:=fCharSureMarker + sValue;
 // Assign
 Guess[fRow,fCol]:=sValue;
 Inc(fRecs);
 fbDataSaved:=False;
end;
// ***************************************************************************//
Function TFileArray.GetElement:String;
// PUBLIC METHOD
// Note - no parameters
Begin
 ErrorString:='TFileArray.GetElement';
 Result:='[Empty]';
 If fRow < 0 then Exit;
 If fCol < 0 then Exit;
 If fRow > High(Guess) then Exit;
 If fCol > High(Guess[fRow]) then Exit;
 Result:=Guess[fRow,fCol];
 If Length(Result) > 1 then
   If Result[1]=fCharSureMarker then
     fbSure:=TRUE
   else
     fbSure:=FALSE;
end;
// ***************************************************************************//
Function TFileArray.GetElementMaxCol:Integer;
begin
 ErrorString:='TFileArray.GetElementMaxCol';
// For a given row (frow)
// Determines number of valid entries
 Result:=0;
 If fRow > High(Guess) then Exit;
 Result:=High(Guess[fRow]);
end;
// ***************************************************************************//
// Property SETs
Procedure TFileArray.SetRow(iValue:Integer);
begin fRow:=iValue; end;
Procedure TFileArray.SetCol(iValue:Integer);
begin fCol:=iValue; end;
Procedure TFileArray.SetUseUnsure(bValue:Boolean);
begin fbUseUnsure:=bValue; end;
Procedure TFileArray.SetSureMarker(charValue:Char);
begin fCharSureMarker:=charValue; end;
Procedure TFileArray.SetSure(bValue:Boolean);
begin fbSure:=bValue; end;
Procedure TFileArray.SetErrorString(aString:String);
begin fErrorString:=aString; end;
Procedure TFileArray.SetDatFilePath(aString:String);
begin fDatFilePath:=aString; end;
// ***************************************************************************//
// ***************************************************************************//
// ***************************************************************************//
// ***************************************************************************//
// ***************************************************************************//
end.
